<?php

namespace Drupal\media_from_images\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Pager\PagerManagerInterface;
use Drupal\Core\Render\Markup;
use Drupal\media_from_images\MediaFromImagesServiceInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Drupal\Core\State\StateInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Admin form for Media From Images module.
 */
class MediaFromImagesAdminForm extends FormBase {

  /**
   * The media from images service.
   *
   * @var \Drupal\media_from_images\MediaFromImagesServiceInterface
   */
  protected MediaFromImagesServiceInterface $mediaFromImagesService;

  /**
   * The pager manager service.
   *
   * @var \Drupal\Core\Pager\PagerManagerInterface
   */
  protected PagerManagerInterface $pagerManager;

  /**
   * @var \Symfony\Component\HttpFoundation\Request|null
   * The current request object.
   */
  protected ?Request $currentRequest;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected StateInterface $state;

  /**
   * Constructs a MediaFromImagesAdminForm object.
   *
   * @param \Drupal\media_from_images\MediaFromImagesServiceInterface $media_from_images_service
   *   The media from images service.
   * @param \Drupal\Core\Pager\PagerManagerInterface $pager_manager
   *   The pager manager service.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request
   *   The request stack.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   */
  public function __construct(
    MediaFromImagesServiceInterface $media_from_images_service,
    PagerManagerInterface $pager_manager,
    RequestStack $request,
    StateInterface $state,
  ) {
    $this->mediaFromImagesService = $media_from_images_service;
    $this->pagerManager = $pager_manager;
    $this->currentRequest = $request->getCurrentRequest();
    $this->state = $state;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): MediaFromImagesAdminForm {
    /** @var \Drupal\media_from_images\MediaFromImagesServiceInterface $mediaFromServices */
    $mediaFromServices = $container->get('media_from_images.service');
    /** @var \Drupal\Core\Pager\PagerManagerInterface $pagerManager */
    $pagerManager = $container->get('pager.manager');
    /** @var \Symfony\Component\HttpFoundation\RequestStack $request */
    $request = $container->get('request_stack');
    /** @var \Drupal\Core\State\StateInterface $state */
    $state = $container->get('state');
    return new static(
      $mediaFromServices,
      $pagerManager,
      $request,
      $state,
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'media_from_images_admin_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    // Check for cron queue status.
    $cron_queue_status = [];

    // Check media creation queue
    $cron_queue_data = $this->state->get('media_from_images_cron_queue');
    if (!empty($cron_queue_data)){
      $cron_queue_status[] = $this->getCronQueueData('images', $cron_queue_data);
    }
    // Check attach queue
    $attach_queue_data = $this->state->get('media_from_images_attach_cron_queue');
    if(!empty($attach_queue_data)){
      $cron_queue_status[] = $this->getCronQueueData('unattached media', $attach_queue_data);
    }
    // Get statistics.
    $total_files = count($this->mediaFromImagesService->getImageFileIds());
    $files_without_media = $this->mediaFromImagesService->countFilesWithoutMediaEntities();
    $unattached_media = $this->mediaFromImagesService->countUnattachedMediaEntities();
    $total_entries = $this->mediaFromImagesService->countMediaHashEntries();
    $failed_entries = $this->mediaFromImagesService->countFailedProcessingEntries();
    $is_batch_running = $this->mediaFromImagesService->isBatchOperationRunning();

    // Set up pagination.
    $items_per_page = 50;
    $current_page = $this->currentRequest->query->get('page', 0);
    $offset = $current_page * $items_per_page;

    // Get paginated entries.
    $entries = $this->mediaFromImagesService->getMediaHashEntries($items_per_page, $offset);

    // Statistics section.
    $form['statistics']['stats'] = [
      '#theme' => 'media_from_images_stats',
      '#data' => [
        'total_files' => $total_files,
        'files_without_media' => $files_without_media,
        'unattached_media' => $unattached_media,
        'total_entries' => $total_entries,
        'failed_entries' => $failed_entries,
        'cron_queue_status' => $cron_queue_status,
      ],
    ];

    // Show warning if operation is running.
    if ($is_batch_running) {
      $form['operation_warning'] = [
        '#type' => 'container',
        '#attributes' => ['class' => ['messages', 'messages--warning']],
        'message' => [
          '#markup' => $this->t('<strong>Warning:</strong> A batch operation is currently running. Please wait for it to complete before starting another operation.'),
        ],
      ];
    }

    // Batch processing section.
    if ($files_without_media > 0) {
      $form['batch_processing'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Batch Processing'),
      ];

      $form['batch_processing']['description'] = [
        '#type' => 'item',
        '#markup' => $this->t('There are @count image files that do not have corresponding media entities. Choose a processing method below.', [
          '@count' => $files_without_media,
        ]),
      ];

      $form['batch_processing']['processing_mode'] = [
        '#type' => 'radios',
        '#title' => $this->t('Processing Mode'),
        '#options' => [
          'immediate' => $this->t('Process all batches immediately (no delays)'),
          'cron' => $this->t('Queue for cron processing (recommended for large datasets)'),
        ],
        '#default_value' => 'immediate',
        '#description' => $this->t("Immediate processing will run all batches sequentially without delays. Cron processing will queue the work to be processed by your site's cron jobs."),
      ];

      $form['batch_processing']['process'] = [
        '#type' => 'submit',
        '#value' => $this->t('Create Media Entities (@count files)', ['@count' => $files_without_media]),
        '#submit' => ['::submitBatchProcess'],
        '#disabled' => $is_batch_running,
      ];
    }

    // Attach unattached media section.
    if ($unattached_media > 0) {
      $form['attach_unattached'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Attach Unattached Media'),
      ];

      $form['attach_unattached']['description'] = [
        '#type' => 'item',
        '#markup' => $this->t('There are @count existing media entities that are not tracked in the hash table. Choose a processing method below.', [
          '@count' => $unattached_media,
        ]),
      ];

      $form['attach_unattached']['attach_processing_mode'] = [
        '#type' => 'radios',
        '#title' => $this->t('Processing Mode'),
        '#options' => [
          'immediate' => $this->t('Process all batches immediately (no delays)'),
          'cron' => $this->t('Queue for cron processing (recommended for large datasets)'),
        ],
        '#default_value' => 'immediate',
        '#description' => $this->t('Immediate processing will run all batches sequentially without delays. Cron processing will queue the work to be processed by your site\'s cron jobs.'),
      ];

      $form['attach_unattached']['attach'] = [
        '#type' => 'submit',
        '#value' => $this->t('Attach Media Entities (@count entities)', ['@count' => $unattached_media]),
        '#submit' => ['::submitAttachUnattached'],
        '#disabled' => $is_batch_running,
      ];
    }

    // Media hash entries table.
    if ($total_entries > 0) {
      $form['entries'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Media Hash Entries (@showing of @total)', [
          '@showing' => count($entries),
          '@total' => $total_entries,
        ]),
      ];

      $header = [
        $this->t('Media ID'),
        $this->t('Media Name'),
        $this->t('File ID'),
        $this->t('Filename'),
        $this->t('File Size'),
        $this->t('Hash (truncated)'),
      ];

      $rows = [];
      foreach ($entries as $entry) {
        // Create media link.
        $media_display = $entry['media_link']
          ? Link::fromTextAndUrl($entry['media_name'], $entry['media_link'])
            ->toString()
          : $entry['media_name'];

        // Create file link to the actual image.
        $file_display = $entry['file_link']
          ? Markup::create('<a href="' . htmlspecialchars($entry['file_link'], ENT_QUOTES, 'UTF-8') . '" target="_blank">' . htmlspecialchars($entry['filename'], ENT_QUOTES, 'UTF-8') . '</a>')
          : htmlspecialchars($entry['filename'], ENT_QUOTES, 'UTF-8');

        $rows[] = [
          $entry['media_id'],
          ['data' => ['#markup' => $media_display]],
          $entry['file_id'],
          ['data' => ['#markup' => $file_display]],
          $entry['filesize'],
          substr($entry['hash'], 0, 16) . '...',
        ];
      }

      $form['entries']['table'] = [
        '#type' => 'table',
        '#header' => $header,
        '#rows' => $rows,
        '#sticky' => TRUE,
        '#empty' => $this->t('No media hash entries found.'),
      ];

      // Set up and add pager.
      if ($total_entries > $items_per_page) {
        $this->pagerManager->createPager($total_entries, $items_per_page);
        $form['entries']['pager'] = [
          '#type' => 'pager',
        ];
      }
    }

    // Failed processing entries section.
    if ($failed_entries > 0) {
      $failed_entries_data = $this->mediaFromImagesService->getFailedProcessingEntries(25, 0);

      $form['failed_entries'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Failed/Skipped Processing (@count entries)', [
          '@count' => $failed_entries,
        ]),
        '#collapsible' => TRUE,
        '#collapsed' => TRUE,
      ];

      $form['failed_entries']['description'] = [
        '#type' => 'item',
        '#markup' => $this->t('Files that failed processing or were skipped. Review these to understand why they could not be processed.'),
      ];

      $failed_header = [
        $this->t('File ID'),
        $this->t('Filename'),
        $this->t('Status'),
        $this->t('Error Message'),
        $this->t('Processed Date'),
      ];

      $failed_rows = [];
      foreach ($failed_entries_data as $entry) {
        $failed_rows[] = [
          $entry['file_id'],
          $entry['filename'],
          ucfirst($entry['status']),
          $entry['error_message'] ?: 'No error message',
          $entry['processed_date'],
        ];
      }

      $form['failed_entries']['table'] = [
        '#type' => 'table',
        '#header' => $failed_header,
        '#rows' => $failed_rows,
        '#empty' => $this->t('No failed processing entries found.'),
      ];

      if ($failed_entries > 25) {
        $form['failed_entries']['note'] = [
          '#type' => 'item',
          '#markup' => $this->t('Showing first 25 of @total failed entries.', [
            '@total' => $failed_entries,
          ]),
        ];
      }
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state): void {
    parent::validateForm($form, $form_state);

    // Check if a batch operation is already running.
    if ($this->mediaFromImagesService->isBatchOperationRunning()) {
      $form_state->setErrorByName('', $this->t('A batch operation is currently running. Please wait for it to complete.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // No specific submit action needed here, as we handle batch processing
    // and attaching unattached media through dedicated submit handlers.
    // This method is required to implement the FormInterface.
  }

  /**
   * Submit handler for batch processing.
   */
  public function submitBatchProcess(array &$form, FormStateInterface $form_state): void {
    $processing_mode = $form_state->getValue('processing_mode', 'immediate');

    if ($processing_mode === 'cron') {
      $this->mediaFromImagesService->startCronBatchProcess();
      $this->messenger()
        ->addStatus($this->t('Media creation process has been queued for cron processing. The jobs will be processed during the next cron runs.'));
    }
    else {
      $this->mediaFromImagesService->startImmediateBatchProcess();
      $this->messenger()
        ->addStatus($this->t('Immediate batch process started. You will be redirected when complete.'));
    }
  }

  /**
   * Submit handler for attaching unattached media entities.
   */
  public function submitAttachUnattached(array &$form, FormStateInterface $form_state): void {
    $processing_mode = $form_state->getValue('attach_processing_mode', 'immediate');

    if ($processing_mode === 'cron') {
      $this->mediaFromImagesService->startCronAttachUnattachedProcess();
      $this->messenger()
        ->addStatus($this->t('Media attachment process has been queued for cron processing. The jobs will be processed during the next cron runs.'));
    }
    else {
      $this->mediaFromImagesService->startImmediateAttachUnattachedProcess();
      $this->messenger()
        ->addStatus($this->t('Immediate attachment process started. You will be redirected when complete.'));
    }
  }

  /**
   * @param string $queue_name
   * @param array|null $cron_queue_data
   *
   * @return array
   */
  public function getCronQueueData(string $queue_name, array $cron_queue_data = NULL): array {
    $output = [];
    if ($cron_queue_data && !empty($cron_queue_data['batches'])) {
      $processed_batches = $cron_queue_data['processed_batches'] ?? 0;
      $total_batches = count($cron_queue_data['batches']);
      $total_remaining = ($total_batches - $processed_batches) * 50;
      $output['queue_name'] = $queue_name;
      $output['processed_batches'] = $processed_batches;
      $output['total_batches'] = $total_batches;
      $output['total_remaining'] = $total_remaining;
    }
    return $output;
  }

}
