<?php

namespace Drupal\media_from_images;

use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Database\Connection;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Lock\LockBackendInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\file\FileInterface;
use Drupal\media\MediaInterface;

/**
 * Service for creating media entities from image files.
 */
class MediaFromImagesService implements MediaFromImagesServiceInterface {

  use StringTranslationTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected LoggerChannelInterface $logger;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected FileSystemInterface $fileSystem;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected Connection $database;

  /**
   * The lock backend service.
   *
   * @var \Drupal\Core\Lock\LockBackendInterface
   */
  protected LockBackendInterface $lock;

  /**
   * The current user account.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected AccountProxyInterface $currentUser;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected StateInterface $state;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected TimeInterface $time;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected MessengerInterface $messenger;

  /**
   * The cache backend service.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected CacheBackendInterface $cache;

  /**
   * Constructs a MediaFromImagesService object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
   *   The logger channel.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection.
   * @param \Drupal\Core\Lock\LockBackendInterface $lock
   *   The lock backend service.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   The current user account.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache backend service.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    LoggerChannelInterface $logger,
    FileSystemInterface $file_system,
    Connection $database,
    LockBackendInterface $lock,
    AccountProxyInterface $current_user,
    StateInterface $state,
    TimeInterface $time,
    MessengerInterface $messenger,
    CacheBackendInterface $cache,
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->logger = $logger;
    $this->fileSystem = $file_system;
    $this->database = $database;
    $this->lock = $lock;
    $this->currentUser = $current_user;
    $this->state = $state;
    $this->time = $time;
    $this->messenger = $messenger;
    $this->cache = $cache;
  }

  /**
   * {@inheritdoc}
   */
  public function getImageFileIds(): array {
    $file_storage = $this->entityTypeManager->getStorage('file');

    $query = $file_storage->getQuery()
      ->accessCheck(FALSE)
      ->condition('filemime', 'image/%', 'LIKE');

    $file_ids = $query->execute();

    return array_values($file_ids);
  }

  /**
   * {@inheritdoc}
   */
  public function getFileHash(FileInterface $file): ?string {
    $file_uri = $file->getFileUri();
    $file_path = $this->fileSystem->realpath($file_uri);

    // Security: Validate file path and ensure it's within expected directories.
    if (!$file_path || !file_exists($file_path)) {
      $this->logger->error('File does not exist: @uri (ID: @id)', [
        '@uri' => $file_uri,
        '@id' => $file->id(),
      ]);
      return NULL;
    }

    // Security: Ensure the file is readable and is actually an image.
    if (!is_readable($file_path)) {
      $this->logger->error('File is not readable: @path (ID: @id)', [
        '@path' => $file_path,
        '@id' => $file->id(),
      ]);
      return NULL;
    }

    // Security: Validate that the file is actually an image by checking MIME.
    $mime_type = $file->getMimeType();
    if (!str_starts_with($mime_type, 'image/')) {
      $this->logger->error('File is not an image: @mime (ID: @id)', [
        '@mime' => $mime_type,
        '@id' => $file->id(),
      ]);
      return NULL;
    }

    return hash_file('sha256', $file_path);
  }

  /**
   * {@inheritdoc}
   */
  public function mediaEntityExistsForFile(FileInterface $file): bool {
    $hash = $this->getFileHash($file);

    if (!$hash) {
      return FALSE;
    }

    // First check the custom tracking table.
    $query = $this->database->select('media_from_images', 'mfi')
      ->fields('mfi', ['media_image_hash'])
      ->condition('media_image_hash', $hash)
      ->range(0, 1);

    $result = $query->execute()->fetchField();

    if (!empty($result)) {
      return TRUE;
    }

    // If not found in tracking table, check for existing media entities.
    $media_storage = $this->entityTypeManager->getStorage('media');
    $media_query = $media_storage->getQuery()
      ->accessCheck(FALSE)
      ->condition('bundle', 'image')
      ->condition('field_media_image', $file->id())
      ->range(0, 1);

    $media_ids = $media_query->execute();

    return !empty($media_ids);
  }

  /**
   * {@inheritdoc}
   */
  public function createMediaFromFile(FileInterface $file): void {
    // Security: Validate that this is actually an image file.
    $mime_type = $file->getMimeType();
    if (!str_starts_with($mime_type, 'image/')) {
      $this->logger->error('Attempted to create media entity for non-image file: @mime (ID: @id)', [
        '@mime' => $mime_type,
        '@id' => $file->id(),
      ]);
      return;
    }

    try {
      $media_storage = $this->entityTypeManager->getStorage('media');

      // Security: Sanitize filename for use as media name.
      $safe_filename = preg_replace('/[^a-zA-Z0-9._-]/', '_', $file->getFilename());
      /** @var MediaInterface $media */
      $media = $media_storage->create([
        'bundle' => 'image',
        'name' => $safe_filename,
        'field_media_image' => [
          'target_id' => $file->id(),
          'alt' => $safe_filename,
        ],
      ]);

      // Save the media entity first to get an ID.
      $media->save();

      // Create hash entry in tracking table after saving.
      $this->createMediaHashEntry($media, $file);

      $this->logger->info('Created media entity @media_id for file @file_id', [
        '@media_id' => $media->id(),
        '@file_id' => $file->id(),
      ]);
    }
    catch (\Exception $e) {
      $this->logger->error('Failed to create media from file @file_id: @message', [
        '@file_id' => $file->id(),
        '@message' => $e->getMessage(),
      ]);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createMediaHashEntry(MediaInterface $media, FileInterface $file): bool {
    $hash = $this->getFileHash($file);

    if (!$hash) {
      $this->logger->error('Cannot create hash entry - unable to generate hash for file ID: @file_id', [
        '@file_id' => $file->id(),
      ]);
      return FALSE;
    }

    try {
      $fields = [
        'media_image_hash' => $hash,
        'media_image_id' => $media->id(),
        'image_entity_id' => $file->id(),
      ];

      // Add new columns if they exist
      $schema = $this->database->schema();
      if ($schema->fieldExists('media_from_images', 'processing_status')) {
        $fields['processing_status'] = 'success';
      }
      if ($schema->fieldExists('media_from_images', 'processed_timestamp')) {
        $fields['processed_timestamp'] = $this->time->getRequestTime();
      }

      $this->database->insert('media_from_images')
        ->fields($fields)
        ->execute();

      // Invalidate the count cache since we added a new media entity.
      $this->cache->delete('media_from_images:count_files_without_media');

      return TRUE;
    }
    catch (\Exception $e) {
      $this->logger->error('Failed to create media hash entry: @message', [
        '@message' => $e->getMessage(),
      ]);
      return FALSE;
    }
  }

  /**
   * Create a failed processing entry in the tracking table.
   *
   * @param \Drupal\file\FileInterface $file
   *   The file entity that failed processing.
   * @param string $status
   *   The processing status (failed, skipped).
   * @param string $error_message
   *   The error message describing why processing failed.
   *
   * @return bool
   *   TRUE if entry was created successfully, FALSE otherwise.
   */
  public function createFailedProcessingEntry(FileInterface $file, string $status, string $error_message = ''): bool {
    // Check if the new columns exist, if not, just log and return
    $schema = $this->database->schema();
    if (!$schema->fieldExists('media_from_images', 'processing_status')) {
      $this->logger->info('Failed processing entry not recorded - schema not updated yet. File @fid, Status: @status, Message: @message', [
        '@fid' => $file->id(),
        '@status' => $status,
        '@message' => $error_message,
      ]);
      return FALSE;
    }

    $hash = $this->getFileHash($file);

    if (!$hash) {
      // If we can't generate a hash, create a placeholder based on file ID
      $hash = 'file_' . $file->id() . '_failed_' . $this->time->getRequestTime();
    }

    try {
      $fields = [
        'media_image_hash' => $hash,
        'media_image_id' => 0, // No media entity created
        'image_entity_id' => $file->id(),
        'processing_status' => $status,
      ];

      // Add optional columns if they exist
      if ($schema->fieldExists('media_from_images', 'error_message')) {
        $fields['error_message'] = $error_message;
      }
      if ($schema->fieldExists('media_from_images', 'processed_timestamp')) {
        $fields['processed_timestamp'] = $this->time->getRequestTime();
      }

      $this->database->insert('media_from_images')
        ->fields($fields)
        ->execute();

      $this->logger->info('Created failed processing entry for file @fid with status @status: @message', [
        '@fid' => $file->id(),
        '@status' => $status,
        '@message' => $error_message,
      ]);

      return TRUE;
    }
    catch (\Exception $e) {
      $this->logger->error('Failed to create failed processing entry: @message', [
        '@message' => $e->getMessage(),
      ]);
      return FALSE;
    }
  }

  /**
   * Create media entity from file ID with error tracking.
   *
   * @param int $file_id
   *   The file entity ID.
   *
   * @return bool
   *   TRUE if media entity was created successfully, FALSE otherwise.
   */
  public function createMediaEntityFromFile(int $file_id): bool {
    try {
      $file_storage = $this->entityTypeManager->getStorage('file');
      /** @var FileInterface $file */
      $file = $file_storage->load($file_id);

      if (!$file) {
        // Create a minimal tracking entry without a file object
        try {
          $hash = 'file_' . $file_id . '_not_found_' . $this->time->getRequestTime();
          $this->database->insert('media_from_images')
            ->fields([
              'media_image_hash' => $hash,
              'media_image_id' => 0,
              'image_entity_id' => $file_id,
              'processing_status' => 'failed',
              'error_message' => 'File entity not found',
              'processed_timestamp' => $this->time->getRequestTime(),
            ])
            ->execute();
        }
        catch (\Exception $e) {
          $this->logger->error('Failed to create tracking entry for missing file @fid: @message', [
            '@fid' => $file_id,
            '@message' => $e->getMessage(),
          ]);
        }
        return FALSE;
      }

      // Check if file URL is accessible
      $file_uri = $file->getFileUri();
      $file_url = $this->fileSystem->realpath($file_uri);

      if (!$file_url || !file_exists($file_url)) {
        $this->createFailedProcessingEntry($file, 'failed', 'File URL returns false or file does not exist');
        return FALSE;
      }

      // Check if media entity already exists
      if ($this->mediaEntityExistsForFile($file)) {
        $this->createFailedProcessingEntry($file, 'skipped', 'Media entity already exists');
        return FALSE;
      }

      // Create the media entity
      $this->createMediaFromFile($file);
      return TRUE;
    }
    catch (\Exception $e) {
      if (isset($file)) {
        $this->createFailedProcessingEntry($file, 'failed', $e->getMessage());
      }
      $this->logger->error('Error creating media entity from file @fid: @message', [
        '@fid' => $file_id,
        '@message' => $e->getMessage(),
      ]);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function startBatchProcess(): void {
    $lock_name = 'media_from_images_create_batch';
    $rate_limit_key = 'media_from_images_rate_limit_' . $this->currentUser->id();

    // Check rate limiting - allow one operation per user per 5 minutes.
    $last_operation = $this->state->get($rate_limit_key, 0);
    $current_time = $this->time->getRequestTime();
    if (($current_time - $last_operation) < 300) {
      $this->logger->warning('Rate limit exceeded for user @uid attempting batch process.', [
        '@uid' => $this->currentUser->id(),
      ]);
      $this->messenger->addWarning($this->t('Please wait at least 5 minutes between batch operations.'));
      return;
    }

    // Try to acquire lock with 30 minute timeout.
    if (!$this->lock->acquire($lock_name, 1800)) {
      $this->logger->warning('Cannot start batch process - another operation is already running.');
      $this->messenger->addWarning($this->t('Another batch operation is currently running. Please wait for it to complete.'));
      return;
    }

    // Set rate limit timestamp.
    $this->state->set($rate_limit_key, $current_time);

    $file_ids = $this->getImageFileIds();

    if (empty($file_ids)) {
      $this->lock->release($lock_name);
      $this->logger->info('No image files found to process.');
      return;
    }

    $operations = [];
    $batch_size = 50;

    // Split file IDs into batches.
    $batches = array_chunk($file_ids, $batch_size);

    foreach ($batches as $batch_file_ids) {
      $operations[] = [
        '\Drupal\media_from_images\MediaFromImagesService::processBatch',
        [$batch_file_ids],
      ];
    }

    $batch = [
      'title' => $this->t('Creating media entities from image files'),
      'operations' => $operations,
      'finished' => '\Drupal\media_from_images\MediaFromImagesService::batchFinished',
      'progress_message' => $this->t('Processing @current of @total batches.'),
    ];

    batch_set($batch);
  }

  /**
   * Batch callback method. Static context requires static service injection.
   *
   * @phpcs:disable DrupalPractice.CodeAnalysis.StaticDrupalApi
   */
  /**
   * Batch process callback to create media entities.
   *
   * @param array $file_ids
   *   Array of file IDs to process.
   * @param array $context
   *   Batch context array.
   */
  public static function processBatch(array $file_ids, array &$context): void {
    $service = \Drupal::service('media_from_images.service');
    $entity_type_manager = \Drupal::entityTypeManager();
    $file_storage = $entity_type_manager->getStorage('file');

    if (!isset($context['sandbox']['processed'])) {
      $context['sandbox']['processed'] = 0;
      $context['sandbox']['created'] = 0;
      $context['sandbox']['skipped'] = 0;
    }

    foreach ($file_ids as $file_id) {
      $file = $file_storage->load($file_id);

      if (!$file) {
        $context['sandbox']['skipped']++;
        continue;
      }

      if (!$service->mediaEntityExistsForFile($file)) {
        $service->createMediaFromFile($file);
        $context['sandbox']['created']++;
      }
      else {
        $context['sandbox']['skipped']++;
      }

      $context['sandbox']['processed']++;
    }

    $context['message'] = t('Processed @processed files. Created @created media entities. Skipped @skipped files.', [
      '@processed' => $context['sandbox']['processed'],
      '@created' => $context['sandbox']['created'],
      '@skipped' => $context['sandbox']['skipped'],
    ]);

    $context['results']['processed'] = $context['sandbox']['processed'];
    $context['results']['created'] = $context['sandbox']['created'];
    $context['results']['skipped'] = $context['sandbox']['skipped'];
  }

  /**
   * Batch finished callback. Static context requires static service injection.
   *
   * @phpcs:disable DrupalPractice.CodeAnalysis.StaticDrupalApi
   */
  /**
   * Batch finished callback.
   *
   * @param bool $success
   *   Indicates whether the batch process was successful.
   * @param array $results
   *   Array of results from the batch process.
   * @param array $operations
   *   Array of operations that were processed.
   */
  public static function batchFinished(bool $success, array $results, array $operations): void {
    $messenger = \Drupal::messenger();
    $lock = \Drupal::service('lock');
    $cache = \Drupal::service('cache.default');

    // Release the lock.
    $lock->release('media_from_images_create_batch');

    // Clear the count cache after batch operations complete.
    $cache->delete('media_from_images:count_files_without_media');

    if ($success) {
      $message = t('Batch process completed successfully. Processed @processed files, created @created media entities, skipped @skipped files.', [
        '@processed' => $results['processed'] ?? 0,
        '@created' => $results['created'] ?? 0,
        '@skipped' => $results['skipped'] ?? 0,
      ]);
      $messenger->addStatus($message);
    }
    else {
      $messenger->addError(t('Batch process finished with errors.'));
    }
  }

  /**
   * Start immediate batch processing (all batches run sequentially without
   * delays).
   */
  public function startImmediateBatchProcess(): void {
    $lock_name = 'media_from_images_create_batch';
    $rate_limit_key = 'media_from_images_rate_limit_' . $this->currentUser->id();

    // Check rate limiting - allow one operation per user per 5 minutes.
    $last_operation = $this->state->get($rate_limit_key, 0);
    $current_time = $this->time->getRequestTime();
    if (($current_time - $last_operation) < 300) {
      $this->logger->warning('Rate limit exceeded for user @uid attempting immediate batch process.', [
        '@uid' => $this->currentUser->id(),
      ]);
      $this->messenger->addWarning($this->t('Please wait at least 5 minutes between batch operations.'));
      return;
    }

    // Try to acquire lock with 2 hour timeout for immediate processing.
    if (!$this->lock->acquire($lock_name, 7200)) {
      $this->logger->warning('Cannot start immediate batch process - another operation is already running.');
      $this->messenger->addWarning($this->t('Another batch operation is currently running. Please wait for it to complete.'));
      return;
    }

    // Set rate limit timestamp.
    $this->state->set($rate_limit_key, $current_time);

    $file_ids = $this->getImageFileIds();

    if (empty($file_ids)) {
      $this->lock->release($lock_name);
      $this->logger->info('No image files found to process.');
      return;
    }

    $operations = [];
    $batch_size = 50;

    // Split file IDs into batches.
    $batches = array_chunk($file_ids, $batch_size);

    foreach ($batches as $batch_file_ids) {
      $operations[] = [
        '\Drupal\media_from_images\MediaFromImagesService::processImmediateBatch',
        [$batch_file_ids],
      ];
    }

    $batch = [
      'title' => $this->t('Creating media entities immediately (no delays)'),
      'operations' => $operations,
      'finished' => '\Drupal\media_from_images\MediaFromImagesService::immediateBatchFinished',
      'progress_message' => $this->t('Processing @current of @total batches immediately.'),
    ];

    batch_set($batch);
  }

  /**
   * Start cron-based batch processing (queued for cron execution).
   */
  public function startCronBatchProcess(): void {
    $rate_limit_key = 'media_from_images_rate_limit_' . $this->currentUser->id();

    // Check rate limiting - allow one operation per user per 5 minutes.
    $last_operation = $this->state->get($rate_limit_key, 0);
    $current_time = $this->time->getRequestTime();
    if (($current_time - $last_operation) < 300) {
      $this->logger->warning('Rate limit exceeded for user @uid attempting cron batch process.', [
        '@uid' => $this->currentUser->id(),
      ]);
      $this->messenger->addWarning($this->t('Please wait at least 5 minutes between batch operations.'));
      return;
    }

    // Set rate limit timestamp.
    $this->state->set($rate_limit_key, $current_time);

    $file_ids = $this->getImageFileIds();

    if (empty($file_ids)) {
      $this->logger->info('No image files found to process.');
      $this->messenger->addMessage($this->t('No image files found that need media entities.'));
      return;
    }

    $batch_size = 50;
    $batches = array_chunk($file_ids, $batch_size);

    // Store batches in queue for cron processing.
    $queue_data = [
      'batches' => $batches,
      'created' => $current_time,
      'user_id' => $this->currentUser->id(),
      'total_files' => count($file_ids),
      'processed_batches' => 0,
    ];

    $this->state->set('media_from_images_cron_queue', $queue_data);

    $this->logger->info('Queued @count batches (@total files) for cron processing.', [
      '@count' => count($batches),
      '@total' => count($file_ids),
    ]);
  }

  /**
   * Process immediate batch callback (processes without delays).
   */
  public static function processImmediateBatch(array $file_ids, array &$context): void {
    $container = \Drupal::getContainer();
    /** @var \Drupal\media_from_images\MediaFromImagesServiceInterface $service */
    $service = $container->get('media_from_images.service');
    /** @var \Drupal\Core\Logger\LoggerChannelInterface $logger */
    $logger = $container->get('logger.channel.media_from_images');

    if (!isset($context['results']['processed'])) {
      $context['results']['processed'] = 0;
      $context['results']['created'] = 0;
      $context['results']['skipped'] = 0;
    }

    foreach ($file_ids as $file_id) {
      try {
        $result = $service->createMediaEntityFromFile($file_id);
        $context['results']['processed']++;

        if ($result) {
          $context['results']['created']++;
        }
        else {
          $context['results']['skipped']++;
        }
      }
      catch (\Exception $e) {
        $logger->error('Error processing file @fid: @message', [
          '@fid' => $file_id,
          '@message' => $e->getMessage(),
        ]);
        $context['results']['skipped']++;
      }
    }

    $context['message'] = t('Processed @processed files. Created @created, skipped @skipped.', [
      '@processed' => $context['results']['processed'],
      '@created' => $context['results']['created'],
      '@skipped' => $context['results']['skipped'],
    ]);
  }

  /**
   * Immediate batch finished callback.
   */
  public static function immediateBatchFinished(bool $success, array $results, array $operations): void {
    $container = \Drupal::getContainer();
    /** @var \Drupal\Core\Lock\LockBackendInterface $lock */
    $lock = $container->get('lock');
    /** @var \Drupal\Core\Messenger\MessengerInterface $messenger */
    $messenger = $container->get('messenger');
    /** @var \Drupal\Core\Cache\CacheBackendInterface $cache */
    $cache = $container->get('cache.default');

    // Release the lock.
    $lock->release('media_from_images_create_batch');

    // Clear the count cache.
    $cache->delete('media_from_images_counts');

    if ($success) {
      $message = t('Immediate processing completed successfully. Processed @processed files, created @created media entities, skipped @skipped files.', [
        '@processed' => $results['processed'] ?? 0,
        '@created' => $results['created'] ?? 0,
        '@skipped' => $results['skipped'] ?? 0,
      ]);
      $messenger->addStatus($message);
    }
    else {
      $messenger->addError(t('Immediate batch process finished with errors.'));
    }
  }

  /**
   * Process cron queue for batch operations.
   */
  public function processCronQueue(): void {
    // Process media creation queue first
    $this->processMediaCreationCronQueue();

    // Then process attach queue
    $this->processAttachCronQueue();
  }

  /**
   * Process media creation cron queue.
   */
  private function processMediaCreationCronQueue(): void {
    $queue_data = $this->state->get('media_from_images_cron_queue');

    if (!$queue_data || empty($queue_data['batches'])) {
      return;
    }

    $lock_name = 'media_from_images_cron_processing';

    // Try to acquire lock with 10 minute timeout.
    if (!$this->lock->acquire($lock_name, 600)) {
      return;
    }

    try {
      $processed_count = 0;
      $max_batches_per_cron = 3; // Process up to 3 batches per cron run

      $batches = $queue_data['batches'];
      $processed_batches = $queue_data['processed_batches'] ?? 0;

      // Process up to max_batches_per_cron batches
      for ($i = $processed_batches; $i < min($processed_batches + $max_batches_per_cron, count($batches)); $i++) {
        $batch_file_ids = $batches[$i];

        foreach ($batch_file_ids as $file_id) {
          try {
            $result = $this->createMediaEntityFromFile($file_id);
            $processed_count++;

            if ($result) {
              $this->logger->debug('Created media entity for file @fid via cron.', ['@fid' => $file_id]);
            }
          }
          catch (\Exception $e) {
            $this->logger->error('Error processing file @fid via cron: @message', [
              '@fid' => $file_id,
              '@message' => $e->getMessage(),
            ]);
          }
        }

        // Update processed batches count
        $queue_data['processed_batches'] = $i + 1;
        $this->state->set('media_from_images_cron_queue', $queue_data);
      }

      // Check if all batches are processed
      if ($queue_data['processed_batches'] >= count($batches)) {
        // All batches processed, clean up
        $this->state->delete('media_from_images_cron_queue');
        $this->cache->delete('media_from_images_counts');

        $this->logger->info('Cron processing completed. Total files processed: @total', [
          '@total' => $queue_data['total_files'],
        ]);
      }
      else {
        $this->logger->info('Cron processed @count files. @remaining batches remaining.', [
          '@count' => $processed_count,
          '@remaining' => count($batches) - $queue_data['processed_batches'],
        ]);
      }
    }
    finally {
      $this->lock->release($lock_name);
    }
  }

  /**
   * Process attach cron queue.
   */
  private function processAttachCronQueue(): void {
    $queue_data = $this->state->get('media_from_images_attach_cron_queue');

    if (!$queue_data || empty($queue_data['batches'])) {
      return;
    }

    $lock_name = 'media_from_images_attach_cron_processing';

    // Try to acquire lock with 10 minute timeout.
    if (!$this->lock->acquire($lock_name, 600)) {
      return;
    }

    try {
      $processed_count = 0;
      $max_batches_per_cron = 3; // Process up to 3 batches per cron run

      $batches = $queue_data['batches'];
      $processed_batches = $queue_data['processed_batches'] ?? 0;

      // Process up to max_batches_per_cron batches
      for ($i = $processed_batches; $i < min($processed_batches + $max_batches_per_cron, count($batches)); $i++) {
        $batch_media_ids = $batches[$i];

        foreach ($batch_media_ids as $media_id) {
          try {
            $media_storage = $this->entityTypeManager->getStorage('media');
            /** @var MediaInterface $media */
            $media = $media_storage->load($media_id);

            if (!$media) {
              continue;
            }

            $field_items = $media->get('field_media_image');
            if ($field_items->isEmpty()) {
              continue;
            }

            $file_id = $field_items->target_id;
            $file_storage = $this->entityTypeManager->getStorage('file');
            /** @var \Drupal\file\FileInterface $file */
            $file = $file_storage->load($file_id);

            if (!$file) {
              continue;
            }

            // Create hash entry.
            if ($this->createMediaHashEntry($media, $file)) {
              $this->logger->debug('Attached media @mid to hash table via cron.', ['@mid' => $media_id]);
            }

            $processed_count++;
          }
          catch (\Exception $e) {
            $this->logger->error('Error processing media @mid via cron attach: @message', [
              '@mid' => $media_id,
              '@message' => $e->getMessage(),
            ]);
          }
        }

        // Update processed batches count
        $queue_data['processed_batches'] = $i + 1;
        $this->state->set('media_from_images_attach_cron_queue', $queue_data);
      }

      // Check if all batches are processed
      if ($queue_data['processed_batches'] >= count($batches)) {
        // All batches processed, clean up
        $this->state->delete('media_from_images_attach_cron_queue');

        $this->logger->info('Cron attach processing completed. Total media processed: @total', [
          '@total' => $queue_data['total_media'],
        ]);
      }
      else {
        $this->logger->info('Cron processed @count media attachments. @remaining batches remaining.', [
          '@count' => $processed_count,
          '@remaining' => count($batches) - $queue_data['processed_batches'],
        ]);
      }
    }
    finally {
      $this->lock->release($lock_name);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getFailedProcessingEntries(int $limit = 50, int $offset = 0): array {
    // Check if the processing_status column exists
    $schema = $this->database->schema();
    if (!$schema->fieldExists('media_from_images', 'processing_status')) {
      return []; // Column doesn't exist yet, return empty array
    }

    try {
      $query = $this->database->select('media_from_images', 'mfi')
        ->fields('mfi', [
          'media_image_hash',
          'image_entity_id',
          'processing_status',
          'error_message',
          'processed_timestamp',
        ])
        ->condition('processing_status', ['failed', 'skipped'], 'IN')
        ->orderBy('processed_timestamp', 'DESC')
        ->range($offset, $limit);

      $results = $query->execute()->fetchAll();
    }
    catch (\Exception $e) {
      $this->logger->error('Error fetching failed processing entries: @message', [
        '@message' => $e->getMessage(),
      ]);
      return [];
    }
    $entries = [];

    foreach ($results as $result) {
      try {
        $file_storage = $this->entityTypeManager->getStorage('file');
        $file = $file_storage->load($result->image_entity_id);

        $entry = [
          'hash' => $result->media_image_hash,
          'file_id' => $result->image_entity_id,
          'status' => $result->processing_status,
          'error_message' => $result->error_message,
          'processed_timestamp' => $result->processed_timestamp,
          'processed_date' => date('Y-m-d H:i:s', $result->processed_timestamp),
        ];

        if ($file) {
          $entry['filename'] = $file->getFilename();
          $entry['file_uri'] = $file->getFileUri();
          $entry['file_size'] = $file->getSize();
        }
        else {
          $entry['filename'] = 'File not found';
          $entry['file_uri'] = '';
          $entry['file_size'] = 0;
        }

        $entries[] = $entry;
      }
      catch (\Exception $e) {
        $this->logger->error('Error loading file for failed entry: @message', [
          '@message' => $e->getMessage(),
        ]);
      }
    }

    return $entries;
  }

  /**
   * {@inheritdoc}
   */
  public function countFailedProcessingEntries(): int {
    // Check if the processing_status column exists
    $schema = $this->database->schema();
    if (!$schema->fieldExists('media_from_images', 'processing_status')) {
      return 0; // Column doesn't exist yet, return 0
    }

    try {
      $query = $this->database->select('media_from_images', 'mfi')
        ->condition('processing_status', ['failed', 'skipped'], 'IN');
      $query->addExpression('COUNT(*)');

      $result = $query->execute();
      return (int) $result->fetchField();
    }
    catch (\Exception $e) {
      $this->logger->error('Error counting failed processing entries: @message', [
        '@message' => $e->getMessage(),
      ]);
      return 0;
    }
  }

  /**
   * {@inheritdoc}
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  public function getMediaHashEntries(int $limit = 50, int $offset = 0): array {
    $query = $this->database->select('media_from_images', 'mfi')
      ->fields('mfi', ['media_image_hash', 'media_image_id', 'image_entity_id'])
      ->orderBy('media_image_id', 'DESC')
      ->range($offset, $limit);

    $results = $query->execute()->fetchAll();

    $entries = [];
    try {
      $media_storage = $this->entityTypeManager->getStorage('media');
    }
    catch (InvalidPluginDefinitionException|PluginNotFoundException $e) {
      $this->logger->error('Media storage plugin not found: @message', ['@message' => $e->getMessage()]);
      // Return empty if media storage is not available.
      return $entries;
    }

    try {
      $file_storage = $this->entityTypeManager->getStorage('file');
    }
    catch (InvalidPluginDefinitionException|PluginNotFoundException $e) {
      $this->logger->error('File storage plugin not found: @message', ['@message' => $e->getMessage()]);
      // Return empty if file storage is not available.
      return $entries;
    }

    foreach ($results as $result) {
      /** @var MediaInterface $media */
      $media = $media_storage->load($result->media_image_id);
      /** @var FileInterface $file */
      $file = $file_storage->load($result->image_entity_id);

      $entries[] = [
        'hash' => $result->media_image_hash,
        'media_id' => $result->media_image_id,
        'media_name' => $media ? $media->getName() : 'Deleted',
        'media_link' => $media ? $media->toUrl('edit-form') : '',
        'file_id' => $result->image_entity_id,
        'filename' => $file ? $file->getFilename() : 'Deleted',
        'file_link' => $file->createFileUrl(),
        'filesize' => $this->formatBytes($file->getSize()),
      ];
    }

    return $entries;
  }

  /**
   * {@inheritdoc}
   */
  public function countMediaHashEntries(): int {
    $query = $this->database->select('media_from_images', 'mfi')
      ->countQuery();

    return (int) $query->execute()->fetchField();
  }

  /**
   * {@inheritdoc}
   */
  public function countFilesWithoutMediaEntities(): int {
    $cache_key = 'media_from_images:count_files_without_media';
    $cached_data = $this->cache->get($cache_key);

    // Return cached result if available and not expired (5 minutes TTL).
    if ($cached_data !== FALSE) {
      return $cached_data->data;
    }

    // Use a more efficient database-only approach.
    $count = $this->countFilesWithoutMediaEntitiesEfficient();

    // Cache the result for 5 minutes.
    $this->cache->set($cache_key, $count, $this->time->getRequestTime() + 300);

    return $count;
  }

  /**
   * Efficiently count files without media entities using database queries only.
   *
   * @return int
   *   The count of files without media entities.
   */
  private function countFilesWithoutMediaEntitiesEfficient(): int {
    // Get all image file IDs that are NOT in the media_from_images table.
    $database = $this->database;

    $query = $database->select('file_managed', 'f');
    $query->leftJoin('media_from_images', 'mfi', 'f.fid = mfi.image_entity_id');
    $query->condition('f.filemime', 'image/%', 'LIKE');
    $query->isNull('mfi.image_entity_id');
    $query->addExpression('COUNT(f.fid)', 'count');

    $result = $query->execute()->fetchField();

    return (int) $result;
  }

  /**
   * {@inheritdoc}
   */
  public function countUnattachedMediaEntities(): int {
    try {
      $untracked_media_ids = $this->getMediaIds();
    }
    catch (InvalidPluginDefinitionException|PluginNotFoundException $e) {
      $this->logger->error('Media storage plugin not found: @message', ['@message' => $e->getMessage()]);
      return 0;
    }

    return count($untracked_media_ids);
  }

  /**
   * {@inheritdoc}
   */
  public function startAttachUnattachedProcess(): void {
    $lock_name = 'media_from_images_attach_batch';
    $rate_limit_key = 'media_from_images_attach_rate_limit_' . $this->currentUser->id();

    // Check rate limiting - allow one operation per user per 5 minutes.
    $last_operation = $this->state->get($rate_limit_key, 0);
    $current_time = $this->time->getRequestTime();
    if (($current_time - $last_operation) < 300) {
      $this->logger->warning('Rate limit exceeded for user @uid attempting attach process.', [
        '@uid' => $this->currentUser->id(),
      ]);
      $this->messenger->addWarning($this->t('Please wait at least 5 minutes between batch operations.'));
      return;
    }

    // Try to acquire lock with 30 minute timeout.
    if (!$this->lock->acquire($lock_name, 1800)) {
      $this->logger->warning('Cannot start attach process - another operation is already running.');
      $this->messenger->addWarning($this->t('Another batch operation is currently running. Please wait for it to complete.'));
      return;
    }

    // Set rate limit timestamp.
    $this->state->set($rate_limit_key, $current_time);

    try {
      $untracked_media_ids = $this->getMediaIds();
    }
    catch (InvalidPluginDefinitionException|PluginNotFoundException $e) {
      $this->logger->error('Media storage plugin not found: @message', ['@message' => $e->getMessage()]);
      $this->lock->release($lock_name);
      return;
    }

    if (empty($untracked_media_ids)) {
      $this->lock->release($lock_name);
      $this->logger->info('No unattached media entities found to process.');
      return;
    }

    $operations = [];
    $batch_size = 50;

    // Split media IDs into batches.
    $batches = array_chunk($untracked_media_ids, $batch_size);

    foreach ($batches as $batch_media_ids) {
      $operations[] = [
        '\Drupal\media_from_images\MediaFromImagesService::processAttachBatch',
        [$batch_media_ids],
      ];
    }

    $batch = [
      'title' => $this->t('Attaching unattached media entities to hash table'),
      'operations' => $operations,
      'finished' => '\Drupal\media_from_images\MediaFromImagesService::attachBatchFinished',
      'progress_message' => $this->t('Processing @current of @total batches.'),
    ];

    batch_set($batch);
  }

  /**
   * Start immediate attach unattached processing (all batches run sequentially
   * without delays).
   */
  public function startImmediateAttachUnattachedProcess(): void {
    $lock_name = 'media_from_images_attach_batch';
    $rate_limit_key = 'media_from_images_attach_rate_limit_' . $this->currentUser->id();

    // Check rate limiting - allow one operation per user per 5 minutes.
    $last_operation = $this->state->get($rate_limit_key, 0);
    $current_time = $this->time->getRequestTime();
    if (($current_time - $last_operation) < 300) {
      $this->logger->warning('Rate limit exceeded for user @uid attempting immediate attach process.', [
        '@uid' => $this->currentUser->id(),
      ]);
      $this->messenger->addWarning($this->t('Please wait at least 5 minutes between batch operations.'));
      return;
    }

    // Try to acquire lock with 2 hour timeout for immediate processing.
    if (!$this->lock->acquire($lock_name, 7200)) {
      $this->logger->warning('Cannot start immediate attach process - another operation is already running.');
      $this->messenger->addWarning($this->t('Another batch operation is currently running. Please wait for it to complete.'));
      return;
    }

    // Set rate limit timestamp.
    $this->state->set($rate_limit_key, $current_time);

    try {
      $untracked_media_ids = $this->getMediaIds();
    }
    catch (InvalidPluginDefinitionException|PluginNotFoundException $e) {
      $this->logger->error('Media storage plugin not found: @message', ['@message' => $e->getMessage()]);
      $this->lock->release($lock_name);
      return;
    }

    if (empty($untracked_media_ids)) {
      $this->lock->release($lock_name);
      $this->logger->info('No unattached media entities found to process.');
      return;
    }

    $operations = [];
    $batch_size = 50;

    // Split media IDs into batches.
    $batches = array_chunk($untracked_media_ids, $batch_size);

    foreach ($batches as $batch_media_ids) {
      $operations[] = [
        '\Drupal\media_from_images\MediaFromImagesService::processImmediateAttachBatch',
        [$batch_media_ids],
      ];
    }

    $batch = [
      'title' => $this->t('Attaching unattached media entities immediately (no delays)'),
      'operations' => $operations,
      'finished' => '\Drupal\media_from_images\MediaFromImagesService::immediateAttachBatchFinished',
      'progress_message' => $this->t('Processing @current of @total batches immediately.'),
    ];

    batch_set($batch);
  }

  /**
   * Start cron-based attach unattached processing (queued for cron execution).
   */
  public function startCronAttachUnattachedProcess(): void {
    $rate_limit_key = 'media_from_images_attach_rate_limit_' . $this->currentUser->id();

    // Check rate limiting - allow one operation per user per 5 minutes.
    $last_operation = $this->state->get($rate_limit_key, 0);
    $current_time = $this->time->getRequestTime();
    if (($current_time - $last_operation) < 300) {
      $this->logger->warning('Rate limit exceeded for user @uid attempting cron attach process.', [
        '@uid' => $this->currentUser->id(),
      ]);
      $this->messenger->addWarning($this->t('Please wait at least 5 minutes between batch operations.'));
      return;
    }

    // Set rate limit timestamp.
    $this->state->set($rate_limit_key, $current_time);

    try {
      $untracked_media_ids = $this->getMediaIds();
    }
    catch (InvalidPluginDefinitionException|PluginNotFoundException $e) {
      $this->logger->error('Media storage plugin not found: @message', ['@message' => $e->getMessage()]);
      return;
    }

    if (empty($untracked_media_ids)) {
      $this->logger->info('No unattached media entities found to process.');
      $this->messenger->addMessage($this->t('No unattached media entities found that need processing.'));
      return;
    }

    $batch_size = 50;
    $batches = array_chunk($untracked_media_ids, $batch_size);

    // Store batches in queue for cron processing.
    $queue_data = [
      'batches' => $batches,
      'created' => $current_time,
      'user_id' => $this->currentUser->id(),
      'total_media' => count($untracked_media_ids),
      'processed_batches' => 0,
    ];

    $this->state->set('media_from_images_attach_cron_queue', $queue_data);

    $this->logger->info('Queued @count attach batches (@total media) for cron processing.', [
      '@count' => count($batches),
      '@total' => count($untracked_media_ids),
    ]);
  }

  /**
   * Batch process callback to attach unattached media entities.
   *
   * @phpcs:disable DrupalPractice.CodeAnalysis.StaticDrupalApi
   *
   * @param array $media_ids
   *   Array of media IDs to process.
   * @param array $context
   *   Batch context array.
   */
  public static function processAttachBatch(array $media_ids, array &$context): void {
    $service = \Drupal::service('media_from_images.service');
    $entity_type_manager = \Drupal::entityTypeManager();
    try {
      $media_storage = $entity_type_manager->getStorage('media');
    }
    catch (InvalidPluginDefinitionException|PluginNotFoundException $e) {
      \Drupal::logger('media_from_images')
        ->error('Media storage plugin not found: @message', ['@message' => $e->getMessage()]);
      return;
    }
    try {
      $file_storage = $entity_type_manager->getStorage('file');
    }
    catch (InvalidPluginDefinitionException|PluginNotFoundException $e) {
      \Drupal::logger('media_from_images')
        ->error('File storage plugin not found: @message', ['@message' => $e->getMessage()]);
      return;
    }

    if (!isset($context['sandbox']['processed'])) {
      $context['sandbox']['processed'] = 0;
      $context['sandbox']['attached'] = 0;
      $context['sandbox']['skipped'] = 0;
    }

    foreach ($media_ids as $media_id) {
      $media = $media_storage->load($media_id);

      if (!$media) {
        $context['sandbox']['skipped']++;
        continue;
      }

      // Get the file from the media entity.
      $file_field = $media->get('field_media_image');
      if ($file_field->isEmpty()) {
        $context['sandbox']['skipped']++;
        continue;
      }

      $file_id = $file_field->target_id;
      $file = $file_storage->load($file_id);

      if (!$file) {
        $context['sandbox']['skipped']++;
        continue;
      }

      // Create hash entry for this media/file pair.
      if ($service->createMediaHashEntry($media, $file)) {
        $context['sandbox']['attached']++;
      }
      else {
        $context['sandbox']['skipped']++;
      }

      $context['sandbox']['processed']++;
    }

    $context['message'] = t('Processed @processed media entities. Attached @attached. Skipped @skipped.', [
      '@processed' => $context['sandbox']['processed'],
      '@attached' => $context['sandbox']['attached'],
      '@skipped' => $context['sandbox']['skipped'],
    ]);

    $context['results']['processed'] = $context['sandbox']['processed'];
    $context['results']['attached'] = $context['sandbox']['attached'];
    $context['results']['skipped'] = $context['sandbox']['skipped'];
  }

  /**
   * Batch finished callback for attaching. Static context requires static
   * service injection.
   *
   * @phpcs:disable DrupalPractice.CodeAnalysis.StaticDrupalApi
   *
   * @param bool $success
   *   Indicates whether the batch process was successful.
   * @param array $results
   *   Array of results from the batch process.
   * @param array $operations
   *   Array of operations that were processed.
   */
  public static function attachBatchFinished(bool $success, array $results, array $operations): void {
    $messenger = \Drupal::messenger();
    $lock = \Drupal::service('lock');

    // Release the lock.
    $lock->release('media_from_images_attach_batch');

    if ($success) {
      $message = t('Attach process completed successfully. Processed @processed media entities, attached @attached, skipped @skipped.', [
        '@processed' => $results['processed'] ?? 0,
        '@attached' => $results['attached'] ?? 0,
        '@skipped' => $results['skipped'] ?? 0,
      ]);
      $messenger->addStatus($message);
    }
    else {
      $messenger->addError(t('Attach process finished with errors.'));
    }
  }

  /**
   * Process immediate attach batch callback (processes without delays).
   */
  public static function processImmediateAttachBatch(array $media_ids, array &$context): void {
    $container = \Drupal::getContainer();
    /** @var \Drupal\media_from_images\MediaFromImagesServiceInterface $service */
    $service = $container->get('media_from_images.service');
    /** @var \Drupal\Core\Logger\LoggerChannelInterface $logger */
    $logger = $container->get('logger.channel.media_from_images');

    if (!isset($context['results']['processed'])) {
      $context['results']['processed'] = 0;
      $context['results']['attached'] = 0;
      $context['results']['skipped'] = 0;
    }

    foreach ($media_ids as $media_id) {
      try {
        $media_storage = \Drupal::entityTypeManager()->getStorage('media');
        $media = $media_storage->load($media_id);

        if (!$media instanceof MediaInterface) {
          $context['results']['skipped']++;
          continue;
        }

        $field_items = $media->get('field_media_image');
        if ($field_items->isEmpty()) {
          $context['results']['skipped']++;
          continue;
        }

        $file_id = $field_items->target_id;
        $file_storage = \Drupal::entityTypeManager()->getStorage('file');
        $file = $file_storage->load($file_id);

        if (!$file instanceof FileInterface) {
          $context['results']['skipped']++;
          continue;
        }

        // Create hash entry.
        if ($service->createMediaHashEntry($media, $file)) {
          $context['results']['attached']++;
        }
        else {
          $context['results']['skipped']++;
        }

        $context['results']['processed']++;
      }
      catch (\Exception $e) {
        $logger->error('Error processing media @mid in immediate attach: @message', [
          '@mid' => $media_id,
          '@message' => $e->getMessage(),
        ]);
        $context['results']['skipped']++;
      }
    }

    $context['message'] = t('Processed @processed media. Attached @attached, skipped @skipped.', [
      '@processed' => $context['results']['processed'],
      '@attached' => $context['results']['attached'],
      '@skipped' => $context['results']['skipped'],
    ]);
  }

  /**
   * Immediate attach batch finished callback.
   */
  public static function immediateAttachBatchFinished(bool $success, array $results, array $operations): void {
    $container = \Drupal::getContainer();
    /** @var \Drupal\Core\Lock\LockBackendInterface $lock */
    $lock = $container->get('lock');
    /** @var \Drupal\Core\Messenger\MessengerInterface $messenger */
    $messenger = $container->get('messenger');

    // Release the lock.
    $lock->release('media_from_images_attach_batch');

    if ($success) {
      $message = t('Immediate attach process completed successfully. Processed @processed media entities, attached @attached, skipped @skipped.', [
        '@processed' => $results['processed'] ?? 0,
        '@attached' => $results['attached'] ?? 0,
        '@skipped' => $results['skipped'] ?? 0,
      ]);
      $messenger->addStatus($message);
    }
    else {
      $messenger->addError(t('Immediate attach process finished with errors.'));
    }
  }

  /**
   * Format bytes into human-readable format.
   *
   * @param int $bytes
   *   The number of bytes.
   * @param int $precision
   *   The number of decimal places.
   *
   * @return string
   *   The formatted string.
   */
  protected function formatBytes(int $bytes, int $precision = 2): string {
    $units = ['B', 'KB', 'MB', 'GB', 'TB'];

    for ($i = 0; $bytes > 1024 && $i < count($units) - 1; $i++) {
      $bytes /= 1024;
    }

    return round($bytes, $precision) . ' ' . $units[$i];
  }

  /**
   * {@inheritdoc}
   */
  public function isBatchOperationRunning(): bool {
    return $this->lock->lockMayBeAvailable('media_from_images_create_batch') === FALSE ||
      $this->lock->lockMayBeAvailable('media_from_images_attach_batch') === FALSE ||
      $this->lock->lockMayBeAvailable('media_from_images_cron_processing') === FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteHashEntryByMediaId(int $media_id): bool {
    try {
      $deleted = $this->database->delete('media_from_images')
        ->condition('media_image_id', $media_id)
        ->execute();

      if ($deleted > 0) {
        $this->logger->info('Deleted hash table entry for media ID @media_id', [
          '@media_id' => $media_id,
        ]);
        return TRUE;
      }

      return FALSE;
    }
    catch (\Exception $e) {
      $this->logger->error('Failed to delete hash table entry for media ID @media_id: @message', [
        '@media_id' => $media_id,
        '@message' => $e->getMessage(),
      ]);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deleteHashEntryByFileId(int $file_id): bool {
    try {
      $deleted = $this->database->delete('media_from_images')
        ->condition('image_entity_id', $file_id)
        ->execute();

      if ($deleted > 0) {
        $this->logger->info('Deleted hash table entry for file ID @file_id', [
          '@file_id' => $file_id,
        ]);
        return TRUE;
      }

      return FALSE;
    }
    catch (\Exception $e) {
      $this->logger->error('Failed to delete hash table entry for file ID @file_id: @message', [
        '@file_id' => $file_id,
        '@message' => $e->getMessage(),
      ]);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getMediaIds(): array {
    try {
      $media_storage = $this->entityTypeManager->getStorage('media');
      $file_storage = $this->entityTypeManager->getStorage('file');
    }
    catch (InvalidPluginDefinitionException|PluginNotFoundException $e) {
      $this->logger->error('Media storage plugin not found: @message', ['@message' => $e->getMessage()]);
      return [];
    }

    $media_query = $media_storage->getQuery()
      ->accessCheck(FALSE)
      ->condition('bundle', 'image');

    $all_media_ids = $media_query->execute();

    // Get file hashes that are already tracked
    $tracked_hashes_query = $this->database->select('media_from_images', 'mfi')
      ->fields('mfi', ['media_image_hash']);
    $tracked_hashes = $tracked_hashes_query->execute()->fetchCol();

    if (empty($tracked_hashes)) {
      return $all_media_ids; // No hashes tracked yet, all media need processing
    }

    $untracked_media_ids = [];

    // Check each media entity to see if its file hash is already tracked
    foreach ($all_media_ids as $media_id) {
      $media = $media_storage->load($media_id);
      if (!$media) {
        continue;
      }

      $field_items = $media->get('field_media_image');
      if ($field_items->isEmpty()) {
        continue;
      }

      $file_id = $field_items->target_id;
      $file = $file_storage->load($file_id);
      if (!$file instanceof FileInterface) {
        continue;
      }

      // Get file hash
      $hash = $this->getFileHash($file);
      if (!$hash) {
        continue;
      }

      // If hash is not already tracked, add this media ID to untracked list
      if (!in_array($hash, $tracked_hashes)) {
        $untracked_media_ids[] = $media_id;
      }
    }

    return $untracked_media_ids;
  }

}
