<?php

namespace Drupal\Tests\media_from_images\Kernel;

use Drupal\KernelTests\KernelTestBase;
use Drupal\Tests\media\Traits\MediaTypeCreationTrait;
use Drupal\Tests\TestFileCreationTrait;
use Drupal\file\Entity\File;
use Drupal\media\Entity\Media;

/**
 * Tests batch processing functionality in MediaFromImagesService.
 *
 * @group media_from_images
 */
class MediaFromImagesBatchTest extends KernelTestBase {

  use MediaTypeCreationTrait;
  use TestFileCreationTrait;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'system',
    'user',
    'file',
    'image',
    'media',
    'media_from_images',
    'field',
  ];

  /**
   * The media from images service.
   *
   * @var \Drupal\media_from_images\MediaFromImagesServiceInterface
   */
  protected $service;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * The lock service.
   *
   * @var \Drupal\Core\Lock\LockBackendInterface
   */
  protected $lock;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installEntitySchema('user');
    $this->installEntitySchema('file');
    $this->installEntitySchema('media');
    $this->installSchema('file', ['file_usage']);
    $this->installSchema('media_from_images', ['media_from_images']);
    $this->installConfig(['system', 'field', 'file', 'image', 'media']);

    // Create image media type.
    $this->createMediaType('image', [
      'id' => 'image',
      'label' => 'Image',
    ]);

    $this->service = $this->container->get('media_from_images.service');
    $this->state = $this->container->get('state');
    $this->lock = $this->container->get('lock');
  }

  /**
   * Tests immediate batch processing functionality.
   */
  public function testImmediateBatchProcess(): void {
    // Create test image files with physical files.
    $files = $this->createTestImageFilesWithContent(3);

    // Verify the service can find the files.
    $file_ids = $this->service->getImageFileIds();
    $this->assertCount(3, $file_ids);

    // Since startImmediateBatchProcess() uses Drupal's batch API which doesn't work in tests,
    // let's test the core functionality by processing files directly.
    foreach ($files as $file) {
      if (!$this->service->mediaEntityExistsForFile($file)) {
        $this->service->createMediaFromFile($file);
      }
    }

    // Verify files are processed correctly.
    $processed_files = $this->getProcessedFiles();
    $this->assertCount(3, $processed_files);

    // Verify media entities were created.
    $media_entities = Media::loadMultiple();
    $this->assertCount(3, $media_entities);
  }

  /**
   * Tests cron batch processing functionality.
   */
  public function testCronBatchProcess(): void {
    // Create test image files (enough for multiple batches).
    $files = $this->createTestImageFilesWithContent(125); // 3 batches of 50 + 25

    // Start cron batch processing.
    $this->service->startCronBatchProcess();

    // Verify queue data is stored correctly.
    $queue_data = $this->state->get('media_from_images_cron_queue');
    $this->assertNotNull($queue_data);
    $this->assertCount(3, $queue_data['batches']); // ceil(125/50) = 3 batches
    $this->assertEquals(125, $queue_data['total_files']);
    $this->assertEquals(0, $queue_data['processed_batches']);

    // Since processCronQueue() uses batch API which doesn't work in tests,
    // process files directly using the service methods.
    $file_ids = $this->service->getImageFileIds();
    foreach ($file_ids as $file_id) {
      $file = \Drupal\file\Entity\File::load($file_id);
      if ($file && !$this->service->mediaEntityExistsForFile($file)) {
        $this->service->createMediaFromFile($file);
      }
    }
    
    // Clear the queue to simulate completion.
    $this->state->delete('media_from_images_cron_queue');

    // Verify media entities were created.
    $media_entities = Media::loadMultiple();
    $this->assertCount(125, $media_entities);
  }

  /**
   * Tests cron queue processing with partial completion.
   */
  public function testCronQueuePartialProcessing(): void {
    // Create enough files for 5 batches.
    $files = $this->createTestImageFilesWithContent(250);

    // Start cron batch processing.
    $this->service->startCronBatchProcess();

    // Verify initial queue state.
    $queue_data = $this->state->get('media_from_images_cron_queue');
    $this->assertCount(5, $queue_data['batches']);
    $this->assertEquals(0, $queue_data['processed_batches']);

    // Simulate processing files directly (since batch API doesn't work in tests).
    $file_ids = $this->service->getImageFileIds();
    foreach ($file_ids as $file_id) {
      $file = \Drupal\file\Entity\File::load($file_id);
      if ($file && !$this->service->mediaEntityExistsForFile($file)) {
        $this->service->createMediaFromFile($file);
      }
    }
    
    // Clear the queue to simulate completion.
    $this->state->delete('media_from_images_cron_queue');
  }

  /**
   * Tests lock handling in batch operations.
   */
  public function testBatchLockHandling(): void {
    // Create test files.
    $files = $this->createTestImageFilesWithContent(2);

    // Test that we can acquire and release the lock.
    $this->assertTrue($this->lock->acquire('media_from_images_create_batch', 30));
    $this->lock->release('media_from_images_create_batch');
    
    // Process files directly since batch API doesn't work in tests.
    foreach ($files as $file) {
      if (!$this->service->mediaEntityExistsForFile($file)) {
        $this->service->createMediaFromFile($file);
      }
    }

    // Files should be processed successfully.
    $media_entities = Media::loadMultiple();
    $this->assertCount(2, $media_entities);
  }

  /**
   * Tests cron lock handling.
   */
  public function testCronLockHandling(): void {
    // Create test files and queue them.
    $files = $this->createTestImageFilesWithContent(50);
    $this->service->startCronBatchProcess();

    // Test that we can acquire and release the cron lock.
    $this->assertTrue($this->lock->acquire('media_from_images_cron_processing', 30));
    $this->lock->release('media_from_images_cron_processing');

    // Verify queue was created.
    $queue_data = $this->state->get('media_from_images_cron_queue');
    $this->assertNotNull($queue_data);
    $this->assertEquals(0, $queue_data['processed_batches']);
    
    // Process files directly since batch API doesn't work in tests.
    $file_ids = $this->service->getImageFileIds();
    foreach ($file_ids as $file_id) {
      $file = \Drupal\file\Entity\File::load($file_id);
      if ($file && !$this->service->mediaEntityExistsForFile($file)) {
        $this->service->createMediaFromFile($file);
      }
    }
    
    // Clear queue to simulate completion.
    $this->state->delete('media_from_images_cron_queue');
  }

  /**
   * Tests batch operation status detection.
   */
  public function testBatchOperationRunningDetection(): void {
    // Test that we can acquire and release all batch operation locks.
    $this->assertTrue($this->lock->acquire('media_from_images_create_batch', 30));
    $this->lock->release('media_from_images_create_batch');
    
    $this->assertTrue($this->lock->acquire('media_from_images_attach_batch', 30));
    $this->lock->release('media_from_images_attach_batch');
    
    $this->assertTrue($this->lock->acquire('media_from_images_cron_processing', 30));
    $this->lock->release('media_from_images_cron_processing');
    
    // Test the service method for detecting batch operations.
    $this->assertFalse($this->service->isBatchOperationRunning());
  }

  /**
   * Tests rate limiting in batch operations.
   */
  public function testBatchRateLimiting(): void {
    // Create test files.
    $files = $this->createTestImageFilesWithContent(2);

    // Set rate limit in recent past.
    $user_id = $this->container->get('current_user')->id();
    $rate_limit_key = 'media_from_images_rate_limit_' . $user_id;
    $this->state->set($rate_limit_key, time() - 200); // 200 seconds ago

    // Test rate limiting by checking if operation would be blocked.
    // Since we can't easily test the startImmediateBatchProcess() method directly,
    // we'll verify the rate limit state exists.
    $rate_limit_time = $this->state->get($rate_limit_key);
    $this->assertNotNull($rate_limit_time);
    $this->assertGreaterThan(time() - 300, $rate_limit_time); // Within last 5 minutes

    // Reset rate limit to allow operation.
    $this->state->set($rate_limit_key, time() - 400); // 400 seconds ago

    // Process files directly since batch API doesn't work in tests.
    foreach ($files as $file) {
      if (!$this->service->mediaEntityExistsForFile($file)) {
        $this->service->createMediaFromFile($file);
      }
    }
    
    $media_entities = Media::loadMultiple();
    $this->assertCount(2, $media_entities);
  }

  /**
   * Tests empty queue handling in cron processing.
   */
  public function testEmptyQueueHandling(): void {
    // Process cron queue when no queue exists.
    $this->service->processCronQueue();

    // Should complete without errors.
    $this->assertTrue(TRUE);

    // Set empty queue data.
    $this->state->set('media_from_images_cron_queue', [
      'batches' => [],
      'total_files' => 0,
      'processed_batches' => 0,
    ]);

    // Process again.
    $this->service->processCronQueue();

    // Should complete without errors.
    $this->assertTrue(TRUE);
  }

  /**
   * Tests immediate attach unattached media processing.
   */
  public function testImmediateAttachUnattachedProcess(): void {
    // Create test files and untracked media entities.
    $files = $this->createTestImageFilesWithContent(3);
    $media_entities = [];
    
    foreach ($files as $file) {
      $media = Media::create([
        'bundle' => 'image',
        'name' => 'Untracked Media',
        'field_media_image' => [
          'target_id' => $file->id(),
          'alt' => 'Test image',
        ],
      ]);
      $media->save();
      $media_entities[] = $media;
    }

    // Verify no entries in hash table initially.
    $initial_count = $this->getHashTableCount();
    $this->assertEquals(0, $initial_count);

    // Process attach directly since batch API doesn't work in tests.
    $unattached_media_ids = $this->service->getMediaIds();
    foreach ($unattached_media_ids as $media_id) {
      $media = Media::load($media_id);
      if ($media && $media->hasField('field_media_image') && !$media->get('field_media_image')->isEmpty()) {
        $file = $media->get('field_media_image')->entity;
        if ($file) {
          $this->service->createMediaHashEntry($media, $file);
        }
      }
    }

    // Verify all media entities are now tracked.
    $final_count = $this->getHashTableCount();
    $this->assertEquals(3, $final_count);
  }

  /**
   * Tests cron attach unattached media processing.
   */
  public function testCronAttachUnattachedProcess(): void {
    // Create test files and untracked media entities (enough for multiple batches).
    $files = $this->createTestImageFilesWithContent(125); // 3 batches of 50 + 25
    $media_entities = [];
    
    foreach ($files as $file) {
      $media = Media::create([
        'bundle' => 'image',
        'name' => 'Untracked Media',
        'field_media_image' => [
          'target_id' => $file->id(),
          'alt' => 'Test image',
        ],
      ]);
      $media->save();
      $media_entities[] = $media;
    }

    // Start cron attach processing.
    $this->service->startCronAttachUnattachedProcess();

    // Verify attach queue data is stored correctly.
    $queue_data = $this->state->get('media_from_images_attach_cron_queue');
    $this->assertNotNull($queue_data);
    $this->assertCount(3, $queue_data['batches']); // ceil(125/50) = 3 batches
    $this->assertEquals(125, $queue_data['total_media']);
    $this->assertEquals(0, $queue_data['processed_batches']);

    // Process attach directly since batch API doesn't work in tests.
    $unattached_media_ids = $this->service->getMediaIds();
    foreach ($unattached_media_ids as $media_id) {
      $media = Media::load($media_id);
      if ($media && $media->hasField('field_media_image') && !$media->get('field_media_image')->isEmpty()) {
        $file = $media->get('field_media_image')->entity;
        if ($file) {
          $this->service->createMediaHashEntry($media, $file);
        }
      }
    }
    
    // Clear the queue to simulate completion.
    $this->state->delete('media_from_images_attach_cron_queue');

    // Verify all media entities are now tracked.
    $final_count = $this->getHashTableCount();
    $this->assertEquals(125, $final_count);
  }

  /**
   * Tests attach lock handling.
   */
  public function testAttachLockHandling(): void {
    // Create test files and media entities.
    $files = $this->createTestImageFilesWithContent(2);
    foreach ($files as $file) {
      $media = Media::create([
        'bundle' => 'image',
        'name' => 'Test Media',
        'field_media_image' => [
          'target_id' => $file->id(),
          'alt' => 'Test image',
        ],
      ]);
      $media->save();
    }

    // Test that we can acquire and release the attach lock.
    $this->assertTrue($this->lock->acquire('media_from_images_attach_batch', 30));
    $this->lock->release('media_from_images_attach_batch');
    
    // Process attach directly since batch API doesn't work in tests.
    $unattached_media_ids = $this->service->getMediaIds();
    foreach ($unattached_media_ids as $media_id) {
      $media = Media::load($media_id);
      if ($media && $media->hasField('field_media_image') && !$media->get('field_media_image')->isEmpty()) {
        $file = $media->get('field_media_image')->entity;
        if ($file) {
          $this->service->createMediaHashEntry($media, $file);
        }
      }
    }

    // Media entities should be attached to hash table.
    $count = $this->getHashTableCount();
    $this->assertEquals(2, $count);
  }

  /**
   * Tests attach rate limiting.
   */
  public function testAttachRateLimiting(): void {
    // Create test files and media entities.
    $files = $this->createTestImageFilesWithContent(2);
    foreach ($files as $file) {
      $media = Media::create([
        'bundle' => 'image',
        'name' => 'Test Media',
        'field_media_image' => [
          'target_id' => $file->id(),
          'alt' => 'Test image',
        ],
      ]);
      $media->save();
    }

    // Set rate limit in recent past.
    $user_id = $this->container->get('current_user')->id();
    $rate_limit_key = 'media_from_images_attach_rate_limit_' . $user_id;
    $this->state->set($rate_limit_key, time() - 200); // 200 seconds ago

    // Test rate limiting by checking if operation would be blocked.
    $rate_limit_time = $this->state->get($rate_limit_key);
    $this->assertNotNull($rate_limit_time);
    $this->assertGreaterThan(time() - 300, $rate_limit_time); // Within last 5 minutes

    // Reset rate limit to allow operation.
    $this->state->set($rate_limit_key, time() - 400); // 400 seconds ago

    // Process attach directly since batch API doesn't work in tests.
    $unattached_media_ids = $this->service->getMediaIds();
    foreach ($unattached_media_ids as $media_id) {
      $media = Media::load($media_id);
      if ($media && $media->hasField('field_media_image') && !$media->get('field_media_image')->isEmpty()) {
        $file = $media->get('field_media_image')->entity;
        if ($file) {
          $this->service->createMediaHashEntry($media, $file);
        }
      }
    }
    
    $count = $this->getHashTableCount();
    $this->assertEquals(2, $count);
  }

  /**
   * Tests processing with failed entries (schema compatibility).
   */
  public function testProcessingWithFailedEntries(): void {
    // Create valid files.
    $files = $this->createTestImageFilesWithContent(2);
    
    // Create a file that will fail processing (missing physical file).
    $invalid_file = File::create([
      'filename' => 'missing.jpg',
      'uri' => 'public://missing.jpg',
      'filemime' => 'image/jpeg',
      'status' => 1,
    ]);
    $invalid_file->save();

    // Process only the valid files directly since batch API doesn't work in tests.
    $processed_count = 0;
    foreach ($files as $file) {
      if (!$this->service->mediaEntityExistsForFile($file)) {
        try {
          $this->service->createMediaFromFile($file);
          $processed_count++;
        } catch (\Exception $e) {
          // Should not happen for valid files
        }
      }
    }

    // Valid files should be processed.
    $this->assertEquals(2, $processed_count);
    
    // Verify media entities were created for valid files.
    $media_entities = Media::loadMultiple();
    $this->assertCount(2, $media_entities);

    // Check if failed entries are tracked (if schema supports it).
    $database = $this->container->get('database');
    $schema = $database->schema();
    
    if ($schema->fieldExists('media_from_images', 'processing_status')) {
      // Create a failed processing entry manually to test the functionality.
      $result = $this->service->createFailedProcessingEntry($invalid_file, 'failed', 'Test processing failure');
      $this->assertTrue($result, 'Failed processing entry should be created successfully.');
      
      $failed_count = $database->select('media_from_images', 'mfi')
        ->condition('processing_status', ['failed', 'skipped'], 'IN')
        ->countQuery()
        ->execute()
        ->fetchField();
      
      $this->assertGreaterThan(0, $failed_count);
    }
  }

  /**
   * Helper method to create test image files.
   */
  protected function createTestImageFiles(int $count): array {
    $files = [];
    
    for ($i = 0; $i < $count; $i++) {
      $file = File::create([
        'filename' => "test_image_{$i}.jpg",
        'uri' => "public://test_image_{$i}.jpg",
        'filemime' => 'image/jpeg',
        'status' => 1,
      ]);
      $file->save();
      $files[] = $file;
    }
    
    return $files;
  }

  /**
   * Helper method to create test image files with actual file content.
   */
  protected function createTestImageFilesWithContent(int $count): array {
    $files = [];
    
    for ($i = 0; $i < $count; $i++) {
      $uri = "public://test_image_{$i}.jpg";
      
      // Create unique content for each file to ensure different hashes.
      $unique_content = "test image content {$i} " . str_repeat('x', $i * 10);
      
      // Create the physical file.
      $directory = dirname($this->container->get('file_system')->realpath($uri));
      $this->container->get('file_system')->prepareDirectory($directory, \Drupal\Core\File\FileSystemInterface::CREATE_DIRECTORY);
      file_put_contents($this->container->get('file_system')->realpath($uri), $unique_content);
      
      // Create the file entity.
      $file = File::create([
        'filename' => "test_image_{$i}.jpg",
        'uri' => $uri,
        'filemime' => 'image/jpeg',
        'status' => 1,
        'filesize' => strlen($unique_content),
      ]);
      $file->save();
      $files[] = $file;
    }
    
    return $files;
  }

  /**
   * Helper method to get processed files from hash table.
   */
  protected function getProcessedFiles(): array {
    $database = $this->container->get('database');
    $query = $database->select('media_from_images', 'mfi')
      ->fields('mfi', ['image_entity_id'])
      ->execute();
    
    return $query->fetchCol();
  }

  /**
   * Helper method to get count of entries in hash table.
   */
  protected function getHashTableCount(): int {
    $database = $this->container->get('database');
    $query = $database->select('media_from_images', 'mfi')
      ->countQuery()
      ->execute();
    
    return (int) $query->fetchField();
  }

}