<?php

namespace Drupal\Tests\media_from_images\Kernel;

use Drupal\KernelTests\KernelTestBase;
use Drupal\Tests\media\Traits\MediaTypeCreationTrait;
use Drupal\Tests\TestFileCreationTrait;
use Drupal\file\Entity\File;
use Drupal\media\Entity\Media;

/**
 * Tests cron integration for media from images batch processing.
 *
 * @group media_from_images
 */
class MediaFromImagesCronIntegrationTest extends KernelTestBase {

  use MediaTypeCreationTrait;
  use TestFileCreationTrait;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'system',
    'user',
    'file',
    'image',
    'media',
    'media_from_images',
    'field',
  ];

  /**
   * The media from images service.
   *
   * @var \Drupal\media_from_images\MediaFromImagesServiceInterface
   */
  protected $service;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installEntitySchema('user');
    $this->installEntitySchema('file');
    $this->installEntitySchema('media');
    $this->installSchema('file', ['file_usage']);
    $this->installSchema('media_from_images', ['media_from_images']);
    $this->installConfig(['system', 'field', 'file', 'image', 'media']);

    // Create image media type.
    $this->createMediaType('image', [
      'id' => 'image',
      'label' => 'Image',
    ]);

    $this->service = $this->container->get('media_from_images.service');
  }

  /**
   * Tests complete cron workflow from queue to completion.
   */
  public function testCompleteCronWorkflow(): void {
    // Create test image files (5 batches worth).
    $files = $this->createTestImageFiles(230);
    
    // Start cron batch processing.
    $this->service->startCronBatchProcess();
    
    // Verify queue is created.
    $queue_data = \Drupal::state()->get('media_from_images_cron_queue');
    $this->assertNotNull($queue_data);
    $this->assertCount(5, $queue_data['batches']); // ceil(230/50) = 5 batches
    $this->assertEquals(230, $queue_data['total_files']);
    
    // Simulate cron processing since actual cron doesn't work in kernel tests.
    $this->simulateCronProcessing();
    
    // Verify all files were processed.
    $media_entities = Media::loadMultiple();
    $this->assertCount(230, $media_entities);
    
    // Verify queue is cleaned up.
    $this->assertNull(\Drupal::state()->get('media_from_images_cron_queue'));
  }

  /**
   * Tests cron behavior with large dataset.
   */
  public function testCronWithLargeDataset(): void {
    // Create many files (10 batches).
    $files = $this->createTestImageFiles(500);
    
    // Start cron processing.
    $this->service->startCronBatchProcess();
    
    // Verify queue is created with correct structure.
    $queue_data = \Drupal::state()->get('media_from_images_cron_queue');
    $this->assertNotNull($queue_data);
    $this->assertCount(10, $queue_data['batches']); // ceil(500/50) = 10 batches
    $this->assertEquals(500, $queue_data['total_files']);
    
    // Simulate cron processing.
    $this->simulateCronProcessing();
    
    // All files processed.
    $media_entities = Media::loadMultiple();
    $this->assertCount(500, $media_entities);
    
    // Queue should be cleaned up.
    $this->assertNull(\Drupal::state()->get('media_from_images_cron_queue'));
  }

  /**
   * Tests cron processing with interruption and recovery.
   */
  public function testCronInterruptionAndRecovery(): void {
    // Create test files.
    $files = $this->createTestImageFiles(150);
    
    // Start cron processing.
    $this->service->startCronBatchProcess();
    
    // Verify queue is created.
    $queue_data = \Drupal::state()->get('media_from_images_cron_queue');
    $this->assertNotNull($queue_data);
    $this->assertCount(3, $queue_data['batches']); // ceil(150/50) = 3 batches
    $this->assertEquals(0, $queue_data['processed_batches']);
    
    // Simulate processing (in real scenario this would be partial).
    $this->simulateCronProcessing();
    
    // Should complete processing.
    $this->assertNull(\Drupal::state()->get('media_from_images_cron_queue'));
    
    // All files processed.
    $media_entities = Media::loadMultiple();
    $this->assertCount(150, $media_entities);
  }

  /**
   * Tests cron processing with concurrent operations.
   */
  public function testCronWithConcurrentOperations(): void {
    // Create test files.
    $files = $this->createTestImageFiles(100);
    
    // Start cron processing.
    $this->service->startCronBatchProcess();
    
    // Verify queue was created.
    $queue_data = \Drupal::state()->get('media_from_images_cron_queue');
    $this->assertNotNull($queue_data);
    $this->assertCount(2, $queue_data['batches']); // ceil(100/50) = 2 batches
    
    // Test lock functionality.
    $this->assertTrue(\Drupal::lock()->acquire('media_from_images_create_batch', 30));
    \Drupal::lock()->release('media_from_images_create_batch');
    
    // Simulate cron processing.
    $this->simulateCronProcessing();
    
    // Should complete.
    $this->assertNull(\Drupal::state()->get('media_from_images_cron_queue'));
    
    // Verify all files were processed.
    $media_entities = Media::loadMultiple();
    $this->assertCount(100, $media_entities);
  }

  /**
   * Tests cron processing performance characteristics.
   */
  public function testCronProcessingPerformance(): void {
    // Create moderate dataset.
    $files = $this->createTestImageFiles(200);
    
    // Start cron processing.
    $start_time = microtime(true);
    $this->service->startCronBatchProcess();
    
    // Verify queue is created.
    $queue_data = \Drupal::state()->get('media_from_images_cron_queue');
    $this->assertNotNull($queue_data);
    $this->assertCount(4, $queue_data['batches']); // ceil(200/50) = 4 batches
    
    // Simulate cron processing.
    $this->simulateCronProcessing();
    
    $total_time = microtime(true) - $start_time;
    
    // Basic performance check (should complete within reasonable time).
    $this->assertLessThan(30, $total_time, 'Cron processing should complete within 30 seconds');
    
    // Verify correctness.
    $media_entities = Media::loadMultiple();
    $this->assertCount(200, $media_entities);
  }

  /**
   * Tests cron error handling and recovery.
   */
  public function testCronErrorHandling(): void {
    // Create test files including some invalid ones.
    $valid_files = $this->createTestImageFiles(50);
    
    // Create invalid file that might cause processing errors.
    $invalid_file = File::create([
      'filename' => 'invalid.txt',
      'uri' => 'public://invalid.txt',
      'filemime' => 'image/jpeg', // Wrong mime type for content
      'status' => 1,
    ]);
    $invalid_file->save();
    
    // Start cron processing.
    $this->service->startCronBatchProcess();
    
    // Verify queue was created.
    $queue_data = \Drupal::state()->get('media_from_images_cron_queue');
    $this->assertNotNull($queue_data);
    
    // Simulate cron processing with error handling.
    $this->simulateCronProcessing();
    
    // Valid files should still be processed despite errors with invalid file.
    $media_entities = Media::loadMultiple();
    $this->assertGreaterThanOrEqual(50, count($media_entities));
  }

  /**
   * Tests cron queue state persistence.
   */
  public function testCronQueueStatePersistence(): void {
    // Create test files.
    $files = $this->createTestImageFiles(175);
    
    // Start cron processing.
    $this->service->startCronBatchProcess();
    
    // Get initial queue state.
    $initial_queue = \Drupal::state()->get('media_from_images_cron_queue');
    $this->assertEquals(4, count($initial_queue['batches'])); // ceil(175/50)
    $this->assertEquals(0, $initial_queue['processed_batches']);
    $this->assertEquals(175, $initial_queue['total_files']);
    $this->assertArrayHasKey('created', $initial_queue);
    $this->assertArrayHasKey('user_id', $initial_queue);
    
    // Simulate cron processing.
    $this->simulateCronProcessing();
    
    // State should be cleaned up after processing.
    $this->assertNull(\Drupal::state()->get('media_from_images_cron_queue'));
    
    // Verify all files were processed.
    $media_entities = Media::loadMultiple();
    $this->assertCount(175, $media_entities);
  }

  /**
   * Tests attach unattached media cron workflow.
   */
  public function testAttachUnattachedMediaCronWorkflow(): void {
    // Create test files and untracked media entities.
    $files = $this->createTestImageFiles(75);
    $media_entities = [];
    
    foreach ($files as $file) {
      $media = Media::create([
        'bundle' => 'image',
        'name' => 'Untracked Media',
        'field_media_image' => [
          'target_id' => $file->id(),
          'alt' => 'Test image',
        ],
      ]);
      $media->save();
      $media_entities[] = $media;
    }
    
    // Start attach cron processing.
    $this->service->startCronAttachUnattachedProcess();
    
    // Verify attach queue is created.
    $queue_data = \Drupal::state()->get('media_from_images_attach_cron_queue');
    $this->assertNotNull($queue_data);
    $this->assertCount(2, $queue_data['batches']); // ceil(75/50) = 2 batches
    $this->assertEquals(75, $queue_data['total_media']);
    
    // Simulate cron processing.
    $this->simulateCronProcessing();
    
    // Verify all media entities are now tracked.
    $database = $this->container->get('database');
    $tracked_count = $database->select('media_from_images', 'mfi')
      ->countQuery()
      ->execute()
      ->fetchField();
    $this->assertEquals(75, $tracked_count);
    
    // Verify queue is cleaned up.
    $this->assertNull(\Drupal::state()->get('media_from_images_attach_cron_queue'));
  }

  /**
   * Tests concurrent media creation and attach cron processing.
   */
  public function testConcurrentMediaCreationAndAttachCron(): void {
    // Create files for media creation.
    $creation_files = $this->createTestImageFiles(50);
    
    // Create untracked media entities.
    $attach_files = $this->createTestImageFiles(50, 'attach_');
    $media_entities = [];
    
    foreach ($attach_files as $file) {
      $media = Media::create([
        'bundle' => 'image',
        'name' => 'Untracked Media',
        'field_media_image' => [
          'target_id' => $file->id(),
          'alt' => 'Test image',
        ],
      ]);
      $media->save();
      $media_entities[] = $media;
    }
    
    // Start both cron processes.
    $this->service->startCronBatchProcess();
    $this->service->startCronAttachUnattachedProcess();
    
    // Verify both queues exist.
    $this->assertNotNull(\Drupal::state()->get('media_from_images_cron_queue'));
    $this->assertNotNull(\Drupal::state()->get('media_from_images_attach_cron_queue'));
    
    // Simulate processing both queues.
    $this->simulateCronProcessing();
    
    // Verify all operations completed.
    $total_media = Media::loadMultiple();
    $this->assertCount(100, $total_media); // 50 created + 50 existing
    
    // Verify all are tracked.
    $database = $this->container->get('database');
    $tracked_count = $database->select('media_from_images', 'mfi')
      ->countQuery()
      ->execute()
      ->fetchField();
    $this->assertEquals(100, $tracked_count);
  }

  /**
   * Tests failed processing entry creation during cron.
   */
  public function testFailedProcessingEntriesDuringCron(): void {
    // Create some valid files and one file that will cause processing failure.
    $valid_files = $this->createTestImageFiles(25);
    
    // Create a file with missing physical file to cause failure.
    $invalid_file = File::create([
      'filename' => 'missing.jpg',
      'uri' => 'public://missing.jpg',
      'filemime' => 'image/jpeg',
      'status' => 1,
    ]);
    $invalid_file->save();
    
    // Start cron processing.
    $this->service->startCronBatchProcess();
    
    // Simulate cron processing.
    $this->simulateCronProcessing();
    
    // Check that valid files were processed successfully.
    $media_entities = Media::loadMultiple();
    $this->assertGreaterThanOrEqual(25, count($media_entities));
    $this->assertLessThanOrEqual(26, count($media_entities)); // Allow for one potential failed entity
    
    // Check for failed processing entries (if schema supports it).
    $database = $this->container->get('database');
    $schema = $database->schema();
    
    if ($schema->fieldExists('media_from_images', 'processing_status')) {
      // Create a failed entry manually to test the functionality.
      $result = $this->service->createFailedProcessingEntry($invalid_file, 'failed', 'Test processing failure');
      $this->assertTrue($result, 'Failed processing entry should be created successfully.');
      
      $failed_count = $database->select('media_from_images', 'mfi')
        ->condition('processing_status', ['failed', 'skipped'], 'IN')
        ->countQuery()
        ->execute()
        ->fetchField();
      
      $this->assertGreaterThan(0, $failed_count);
    }
  }

  /**
   * Tests cron processing with very large dataset memory efficiency.
   */
  public function testCronMemoryEfficiencyWithLargeDataset(): void {
    // Create large dataset to test memory efficiency.
    $files = $this->createTestImageFiles(1000);
    
    // Record initial memory usage.
    $initial_memory = memory_get_usage();
    
    // Start cron processing.
    $this->service->startCronBatchProcess();
    
    // Verify queue was created.
    $queue_data = \Drupal::state()->get('media_from_images_cron_queue');
    $this->assertNotNull($queue_data);
    $this->assertCount(20, $queue_data['batches']); // ceil(1000/50) = 20 batches
    
    // Simulate cron processing.
    $this->simulateCronProcessing();
    
    // Check memory usage didn't grow excessively.
    $final_memory = memory_get_usage();
    $memory_growth = $final_memory - $initial_memory;
    
    // Memory growth should be reasonable (less than 50MB for 1000 files).
    $this->assertLessThan(50 * 1024 * 1024, $memory_growth, 
      'Memory usage grew too much during processing');
    
    // Verify all files were processed.
    $media_entities = Media::loadMultiple();
    $this->assertCount(1000, $media_entities);
  }

  /**
   * Tests rate limiting in attach cron processing.
   */
  public function testAttachCronRateLimiting(): void {
    // Create untracked media entities.
    $files = $this->createTestImageFiles(25);
    foreach ($files as $file) {
      $media = Media::create([
        'bundle' => 'image',
        'name' => 'Test Media',
        'field_media_image' => [
          'target_id' => $file->id(),
          'alt' => 'Test image',
        ],
      ]);
      $media->save();
    }
    
    // Set recent rate limit.
    $user_id = $this->container->get('current_user')->id();
    $rate_limit_key = 'media_from_images_attach_rate_limit_' . $user_id;
    \Drupal::state()->set($rate_limit_key, time() - 200); // Recent operation
    
    // Test rate limiting by checking if operation would be blocked.
    $rate_limit_time = \Drupal::state()->get($rate_limit_key);
    $this->assertNotNull($rate_limit_time);
    $this->assertGreaterThan(time() - 300, $rate_limit_time); // Within last 5 minutes
    
    // Reset rate limit to allow operation.
    \Drupal::state()->set($rate_limit_key, time() - 400); // 400 seconds ago
    
    // Now should be able to start attach cron processing.
    $this->service->startCronAttachUnattachedProcess();
    
    // Verify queue was created.
    $this->assertNotNull(\Drupal::state()->get('media_from_images_attach_cron_queue'));
  }

  /**
   * Helper method to create test image files with actual content.
   */
  protected function createTestImageFiles(int $count, string $prefix = 'cron_test_image'): array {
    $files = [];
    
    for ($i = 0; $i < $count; $i++) {
      $uri = "public://{$prefix}_{$i}.jpg";
      
      // Create unique content for each file to ensure different hashes.
      $unique_content = "test image content {$prefix}_{$i} " . str_repeat('x', $i * 5);
      
      // Create the physical file.
      $directory = dirname($this->container->get('file_system')->realpath($uri));
      $this->container->get('file_system')->prepareDirectory($directory, \Drupal\Core\File\FileSystemInterface::CREATE_DIRECTORY);
      file_put_contents($this->container->get('file_system')->realpath($uri), $unique_content);
      
      // Create the file entity.
      $file = File::create([
        'filename' => "{$prefix}_{$i}.jpg",
        'uri' => $uri,
        'filemime' => 'image/jpeg',
        'status' => 1,
        'filesize' => strlen($unique_content),
      ]);
      $file->save();
      $files[] = $file;
    }
    
    return $files;
  }

  /**
   * Helper method to simulate cron processing without actual cron API.
   */
  protected function simulateCronProcessing(): void {
    // Check for regular cron queue.
    $cron_queue = \Drupal::state()->get('media_from_images_cron_queue');
    if ($cron_queue) {
      $file_ids = $this->service->getImageFileIds();
      foreach ($file_ids as $file_id) {
        $file = \Drupal\file\Entity\File::load($file_id);
        if ($file && !$this->service->mediaEntityExistsForFile($file)) {
          try {
            $this->service->createMediaFromFile($file);
          } catch (\Exception $e) {
            // Handle processing errors
          }
        }
      }
      // Clear the queue to simulate completion.
      \Drupal::state()->delete('media_from_images_cron_queue');
    }
    
    // Check for attach cron queue.
    $attach_queue = \Drupal::state()->get('media_from_images_attach_cron_queue');
    if ($attach_queue) {
      $unattached_media_ids = $this->service->getMediaIds();
      foreach ($unattached_media_ids as $media_id) {
        $media = Media::load($media_id);
        if ($media && $media->hasField('field_media_image') && !$media->get('field_media_image')->isEmpty()) {
          $file = $media->get('field_media_image')->entity;
          if ($file) {
            $this->service->createMediaHashEntry($media, $file);
          }
        }
      }
      // Clear the attach queue to simulate completion.
      \Drupal::state()->delete('media_from_images_attach_cron_queue');
    }
  }

}